/*	Interfejs do bazy logw ekg.
 *  Copyright (C) 2004 Piotr Kalinowski <pitkali@o2.pl>
 *  All rights reserved.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#ifndef _LIBEKGLOG_
#define _LIBEKGLOG_

#include <list>
#include <map>
#include <ctime>
#include <stdexcept>
using std::time_t;

/**
 *  @brief Przestrze nazw zawierajca interfejs do odczytu logw ekg.
 */
namespace EkgLog {

  /**
   *  @brief Typ wiadomoci.
   *
   *  Typem wiadomoci moe by:
   *  - MSG_RECV - wiadomo odebrana
   *  - MSG_SENT - wiadomo wysana
   */
  enum msg_type { MSG_RECV, MSG_SENT };

  /**
   *  @brief Klasa wiadomoci.
   *
   *  Ta klasa reprezentuje informacje o pojedynczej wiadomoci w logach ekg.
   *  Nota bene: jako konwencj przyjto, e dla wiadomoci wysanych, czas
   *  odebrania wiadomoci jest rwny czasowi wysania. (czas odebrania pierwszej
   *  wiadomoci okrela czas pocztku rozmowy)
   */
  class Message {
    private:
      time_t _rtime; // Czas otrzymania wiadomoci
      time_t _stime; // Czas wysania wiadomoci
      std::string _msg; // tre
      msg_type _type; // wysana czy odebrana?
      unsigned int _no; // numer w danej rozmowie
    public:
      /// Domylny konstruktor
      Message():_rtime(0),_stime(0),_type(MSG_RECV),_no(0) {}
      /**
       *  @brief Konstruktor
       *  @param recv Czas otrzymania wiadomoci (dla odebranych powinien
       *  by rwny czasowi wysania).
       *  @param sent Czas wysania wiadomoci.
       *  @param s Tre wiadomoci.
       *  @param type Typ wiadomoci.
       *  @param no Numer wiadomoci w danej rozmowie (domylnie 0).
       */
      Message(const time_t recv, const time_t sent,
	  const std::string& s, const msg_type type, unsigned int no = 0)
	: _rtime(recv),_stime(sent),_msg(s),_type(type),_no(no) {}
      /// Destruktor
      ~Message() { }

      // --- dostp do odczytu ---

      /// Pobierz czas otrzymania wiadomoci.
      time_t recv(void) const { return _rtime; }
      /// Pobierz czas wysania wiadomoci.
      time_t sent(void) const { return _stime; }
      /// Pobierz tre wiadomoci.
      const std::string& msg(void) const { return _msg; }
      /**
       *  @brief Pobierz typ wiadomoci.
       *  @see msg_type
       */
      msg_type type(void) const { return _type; }
      /// Pobierz numer wiadomoci w danej rozmowie.
      unsigned int get_no(void) const { return _no; }

      // --- dostp do zapisu ---

      /// Ustaw czas otrzymania wiadomoci.
      void recv(const time_t t) { _rtime = t; }
      /// Ustaw czas wysania wiadomoci (dla wysyanych take odebrania).
      void sent(const time_t t) { _stime = t; if (_type==MSG_SENT) _rtime = t; }
      /// Ustaw tre wiadomoci.
      void msg(const std::string& s) { _msg = s; }
      /**
       *  @brief Inny sposb ustawienia treci wiadomoci.
       *  @param str rdowy string.
       *  @param idx Indeks pocztkowy.
       *
       *  Ustaw tre wiadomoci na zawarto string, ale dopiero od indeksu pocztkowego.
       */
      void msg(const std::string& str, const std::string::size_type idx) { _msg.assign(str,idx,str.size()-idx); }
      /**
       *  @brief Jeszcze inny sposb na ustawienie treci wiadomoci :)
       *  @param str rdowy string.
       *  @param beg_idx Indeks pocztkowy.
       *  @param end_idx Indeks za-kocowy.
       */
      void msg(const std::string& str, const std::string::size_type beg_idx, const std::string::size_type end_idx)
      { _msg.assign(str, beg_idx, end_idx - beg_idx); }
      /// Dodaj string do treci wiadomoci
      void append(const std::string& str);
      /// Ustaw typ wiadomoci.
      void type(const msg_type t) { _type = t; }
      /// Ustaw numer wiadomoci w rozmowie.
      void set_no(const unsigned int n) { _no = n; }
  }; // class Message

  /**
   *  @brief Rodzaj statusu gg.
   *
   *  Moliwe wartoci to:
   *  - AVAILABLE - dostpny
   *  - BUSY - zajty
   *  - INVISIBLE - niewidoczny
   *  - NOTAVAILABLE - niedostpny
   */
  enum status_type { AVAILABLE, BUSY, INVISIBLE, NOTAVAILABLE };

  /**
   *  @brief Klasa reprezentujca zmian statusu.
   */
  class Status {
    private:
      time_t _time; // czas zmiany
      status_type _type; // typ nowego statusu
      std::string _desc; // nowy opis
      unsigned int _no; // numer zmiany statusu danej osoby
    public:
      /// Domylny konstruktor.
      Status():_time(0),_type(NOTAVAILABLE),_no(0) {}
      /**
       *  @brief Konstruktor
       *  @param time Czas zmiany statusu.
       *  @param type Typ nowego statusu.
       *  @param s Nowy opis.
       *  @param no Numer zmiany statusu dla danej osoby.
       */
      Status(const time_t time, const status_type type, const std::string& s,
	  const unsigned int no = 0)
	: _time(time),_type(type),_desc(s),_no(no) {}
      /// Destruktor
      ~Status() {}

      // --- dostp do odczytu ---

      /// Pobierz czas zmiany statusu.
      time_t time(void) const { return _time; }
      /// Pobierz typ nowego statusu.
      status_type type(void) const { return _type; }
      /// Pobierz nowy opis.
      const std::string& desc(void) const { return _desc; }
      /// Pobierz numer zmiany statusu danej osoby.
      unsigned int get_no(void) const { return _no; }

      // --- dostp do zapisu ---

      /// Ustaw czas zmiany statusu.
      void time(const time_t t) { _time = t; }
      /// Ustaw typ nowego statusu.
      void type(const status_type t) { _type = t; }
      /// Ustaw nowy opis.
      void desc(const std::string& s) { _desc = s; }
      /// Ustaw nowy opis.
      void desc(const std::string& s, const std::string::size_type idx) { _desc.assign (s,idx,s.size()-idx); }
      /// Ustaw nowy opis.
      void desc(const std::string& s, const std::string::size_type beg_idx,
	  const std::string::size_type end_idx)
      { _desc.assign (s,beg_idx, end_idx-beg_idx); }
      /// Ustaw nowy numer zmiany statusu danej osoby.
      void set_no(const unsigned int n) { _no = n; }
  }; // class Status

  /// Iterator po licie wiadomoci.
  typedef std::list<Message>::iterator msg_iterator;
  /// Iterator po licie wiadomoci (tylko odczyt).
  typedef std::list<Message>::const_iterator const_msg_iterator;
  /// Iterator po licie statusw.
  typedef std::list<Status>::iterator status_iterator;
  /// Iterator po licie statusw (tylko odczyt).
  typedef std::list<Status>::const_iterator const_status_iterator;

  /**
   *  @brief Klasa przechowujca wynik wyszukiwania.  
   */
  class SearchResult {
    private:
      msg_iterator _msg;
      std::string::size_type _beg, _len;
    public:
      SearchResult (const msg_iterator& m, const std::string::size_type& b, const std::string::size_type& l)
	:_msg(m),_beg(b),_len(l)
	{}
      /**
       *  @brief Zwraca iterator wiadomoci.
       *
       *  W tej wiadomoci wystpio szukane wyraenie. Jeli przeszukiwanie
       *  si nie powiodo, warto ta jest wartoci chat::end().
       */
      const msg_iterator& message (void) const { return _msg; }
      /** 
       *  @brief Zwraca indeks pocztku dopasowania.
       *  @return W przypadku nie znalezienia dopasowania, zwraca string::npos.
       */
      const std::string::size_type& begin (void) const { return _beg; }
      /**
       *  @brief Zwraca dugo dopasowania.
       *  @return W przypadku nie znalezienia dopasowania, zwraca string::npos.
       */
      const std::string::size_type& length (void) const { return _len; }
      /// alias dla SearchResult::length
      const std::string::size_type& size (void) const { return length(); }
  };

  /**
   *  @brief Klasa reprezentujca pojedyncz rozmow.
   *
   *  W zasadzie obiekty tej klasy przechowuj zakresy wiadomoci z bazy
   *  reprezentujce poszczeglne rozmowy. Zakresy s w stylu biblioteki
   *  standardowej, czyli potwarte [pierwszy, ostatni).
   *
   *  Elementy msg_iterator to prawdziwe iteratory, co ozancza, e mona
   *  wobec nich stosowa semantyk wskanikw. Jednak nie s to iteratory
   *  dostpu swobodnego! (nie wspieraj arytmetyki wskanikw).
   */
  class Chat {
    private:
      msg_iterator _begin,_end; // pocztek i koniec rozmowy
    public:
      // brak domylnych konstruktorw i destruktora - niech kompilator sam
      // wygeneruje - to wystarczy...

      /// Pobierz pozycj pierwszej wiadomoci.
      const msg_iterator& begin(void) const { return _begin; }
      /// Pobierz pozycj koca - to nie jest poprawna pozycja w obrbie rozmowy!
      const msg_iterator& end(void) const { return _end; }
      /// Pobierz czas rozpoczcia rozmowy (begin() musi by poprawnym iteratorem!).
      time_t time(void) const { return _begin->recv(); }
      /// Pobierz ilo wiadomoi w czacie
      unsigned int size(void) const { return std::list<Message>::reverse_iterator(_end)->get_no(); }
      /**
       *  @brief Szukaj wyraenia regularnego w rozmowie.
       *  @param regex Wyraenie regularnego, ktrego szukamy.
       *  @param ignore_case Okrela czy przeszukiwanie ma bra pod uwag wielko znakw.
       *  Domylnie: tak.
       *  @return Zwraca obiekt typu SearchResult.
       *
       *  Niniejsza funkcja pobiera wyraenie regularne i szuka jego wystpie w danym czacie.
       *  Moe ona zwrci runtime_error, gdy wystpi wewntrzny bd biblioteki obsugi wyrae
       *  regularnych lub zostanie podane nieprawidowe wyraenie regularne.
       *  Wwczas std::runtime_error::what() zawiera dokadniejsze informacje
       *  w postaci czytelnej dla czowieka.
       */
      SearchResult search (const std::string& regex, const bool ignore_case = false) throw(std::runtime_error);
      /**
       *  @brief Szukaj wyraenia regularnego poczwszy od danej pozycji.
       *  @param regex Wyraenie regularne, ktrego szukamy.
       *  @param prev Pozycja poprzedniego dopasowania.
       *  @param ignore_case Czy bra pod uwag wielko znakw? Domylnie tak.
       *  @see Zobacz Chat::search(const std::string&,const bool).
       */
      SearchResult search (const std::string& regex, const SearchResult& prev, const bool ignore_case = false) throw(std::runtime_error);
      /// Ustaw pocztek rozmowy.
      void begin(const msg_iterator& b) { _begin = b; }
      /// Ustaw koniec rozmowy.
      void end(const msg_iterator& e) { _end = e; }
  };

  /// Numerek uin.
  typedef unsigned int uin_t;
  /// Iterator po mapie rozmw
  typedef std::map<const time_t, Chat>::iterator chat_iterator;
  /// Iterator po mapie rozmw (tylko odczyt).
  typedef std::map<const time_t, Chat>::const_iterator const_chat_iterator;
  /// Iterator odwrotny po mapie rozmw.
  typedef std::map<const time_t, Chat>::reverse_iterator chat_reverse_iterator;
  /// Iterator po mapie rozmw (tylko odczyt).
  typedef std::map<const time_t, Chat>::const_reverse_iterator const_chat_reverse_iterator;

  /**
   *  @brief Klasa reprezentujca cao logw zwizanych z danym numerkiem uin.
   */
  class UinLog {
    private:
      std::map<const time_t, Chat> _chats;
      std::list<Message> _msg;
      std::list<Status> _status;
      uin_t _uin;
      std::string _nick;
    public:
      /// Pozycja pocztkowa listy czatw
      chat_iterator begin(void) { return _chats.begin(); }
      /// Pozycja pocztkowa listy czatw (tryb tylko do odczytu).
      const_chat_iterator begin(void) const { return _chats.begin(); }
      /// Pozycja za-kocowa - jeli wyszukiwanie j zwrci, tzn. e si nie powiodo.
      chat_iterator end(void) { return _chats.end(); }
      /// Pozycja za-kocowa - tryb tylko do odczytu.
      const_chat_iterator end(void) const { return _chats.end(); }

      /// Kierunek odwrotny: Pozycja pocztkowa listy czatw
      chat_reverse_iterator rbegin(void) { return _chats.rbegin(); }
      /// Kierunek odwrotny: Pozycja pocztkowa listy czatw (tryb tylko do odczytu).
      const_chat_reverse_iterator rbegin(void) const { return _chats.rbegin(); }
      /// Kierunek odwrotny: Pozycja za-kocowa - jeli wyszukiwanie j zwrci, tzn. e si nie powiodo.
      chat_reverse_iterator rend(void) { return _chats.rend(); }
      /// Kierunek odwrotny: Pozycja za-kocowa - tryb tylko do odczytu.
      const_chat_reverse_iterator rend(void) const { return _chats.rend(); }

      /// Ile czatw przechowujemy?
      unsigned int size(void) const { return _chats.size(); }

      // --- interfejs do odczytu --- chaty ---
      /// Pobierz nick tego uina.
      const std::string& nick(void) const { return _nick; }
      /// Pobierz uin tego logu.
      uin_t uin(void) const { return _uin; }
      /**
       *  @brief Znajd rozmow o danym czasie rozpoczcia.
       *  @param t Czas rozpoczcia rozmowy.
       *  @return Iterator szukanej rozmowy lub UinLog::end() w przypadku
       *  poraki przy wyszukiwaniu.
       */
      chat_iterator chat(const time_t t) { return _chats.find(t); }
      /** 
       *  @brief Znajd rozmow o danym czasie rozpoczcia. (tylko odczyt)
       *  @param t Czas rozpoczcia rozmowy.
       *  @see Zobacz chat(const time_t).
       *
       *  Funkcja dziaa tak samo jak chat(const time_t) ale dla bazy w trybie
       *  tylko do odczytu.
       */
      const_chat_iterator chat(const time_t t) const { return _chats.find(t); }

      // --- interfejs statusw ---
      /// Pobierz iterator pierwszego statusu.
      status_iterator statusBegin(void) { return _status.begin(); }
      /// Iterator pierwszego statusu dla bazy tylko do odczytu.
      const_status_iterator statusBegin(void) const { return _status.begin(); }
      /// Pobierz iterator koca listy stutusw (nie jest poprawn pozycj!)
      status_iterator statusEnd(void) { return _status.end(); }
      /// Iterator za-kocowy listy statusw (tylko do odczytu).
      const_status_iterator statusEnd(void) const { return _status.end(); }
      /// Pobierz ilo przechowywanych statusw
      unsigned int statusCount (void) const { return _status.size(); }

      // --- lista wiadomoci ---
      /// Pobierz iterator pierwszej wiadomoci.
      msg_iterator msgBegin(void) { return _msg.begin(); }
      /// Iterator pierwsze wiadomoci dla bazy tylko do odczytu.
      const_msg_iterator msgBegin(void) const { return _msg.begin(); }
      /// Pobierz iterator koca listy wiadomoci (nie jest poprawn pozycj!)
      msg_iterator msgEnd(void) { return _msg.end(); }
      /// Iterator za-kocowy listy wiadomoci dla bazy tylko do odczytu.
      const_msg_iterator msgEnd(void) const { return _msg.end(); }
      /// Pobierz ilo przechowywanych wiadomoci
      unsigned int msgCount(void) const { return _msg.size(); }

      // ---------- interfejs do zapisu -----------
      /**
       *  @brief Konstruktor.
       *  @param uin Numerek uin.
       *  @param nick Nick danego numerka.
       */
      UinLog (const uin_t uin = 0, const std::string& nick = ""):_uin(uin),_nick(nick) {}
      /// Ustaw nick.
      void nick(const std::string& s) { _nick = s; }
      /// Ustaw uin.
      void uin (const uin_t u) { _uin = u; }
      /// Dodaj status do listy.
      void addStatus(const Status& s) { _status.push_back(s); }
      /// Usu status.
      void delStatus(const status_iterator& i) 
      { _status.erase (i); }
      /**
       *  @brief Dodaj zakres statusw do listy statusw.
       *  @param beg Iterator pocztka zakresu.
       *  @param end Iterator koca zakresu.
       *  
       *  Funkcja dodaje zakres [begin,end) do listy statusw. Jest to funkcja
       *  do interakcji z bibliotek standardow i zgodnie z jej konwencjami
       *  uytkownik musi zapewni, e podany zakres jest poprawny. Iteratory
       *  musz by iteratorami wejciowymi (wymagane przez list::push_back).
       */
      template<class InputIterator>
	void addStatus(const InputIterator& beg, const InputIterator& end)
	{ copy(beg, end, back_inserter(_status)); }
      /// Dodaj wiadomo do listy.
      void addMessage(const Message& m) { _msg.push_back(m); }
      /// Usu wiadomo.
      void delMessage(const msg_iterator& i) { _msg.erase(i); }
      /**
       *  @brief Dodaj zakres wiadomoci do listy - dziaa jak
       *  addStatus(const InputIterator&,const InputIterator&).
       */
      template<class InputIterator>
	void addMessage(const InputIterator& beg, const InputIterator& end)
	{ copy(beg, end, back_inserter(_msg)); }
      /**
       *  @brief Opracuj podzia na rozmowy.
       *  @param interval Minimalny czas dzielcy dwie rozmowy.
       *
       *  Dzieli list wiadomoci na rozmowy. Miejsca podziau znajduj si tam,
       *  gdzie kolejna wiadomo jest oddalona od poprzedniej w czasie o co
       *  najmniej interval.
       */
      void fillChats(const time_t interval);
      /// Usu dany czat.
      void delChat(const chat_iterator& i) {
	for (msg_iterator m = i->second.begin(); m != i->second.end(); ++m)
	  delMessage(m);
      }
  }; // class UinLog

  /// Iterator po logach w bazie (log_iterator.second dziaa jak wskanik na UinLog).
  typedef std::map<const uin_t, UinLog>::iterator log_iterator;
  /// Iterator po logach w bazie tylko do odczytu.
  typedef std::map<const uin_t, UinLog>::const_iterator const_log_iterator;
  /// Mapowanie uinw do nickw
  typedef std::map<const uin_t, std::string> userlist_t;

  /**
   *  @brief Klasa reprezentujca cao logw ekg na dysku.
   *
   *  To wanie za pomoc tej klasy powinien odbywa si dostp do logw ekg
   *  z poziomu oprogramowania. Ona wczytuje dane z dysku i udostpnia wiadomoci.
   */
  class dBase {
    private:
      std::map<const uin_t, UinLog> _logs;
      userlist_t _nicks;
      bool _badbit;
      // statystyki
      unsigned int _messages, _statuses; 
      double _time;
      // zabro tych operacji
      dBase(const dBase&);
      dBase& operator=(const dBase&);
    public:
      /// Typ reprezentujcy format logw.
      enum log_format_t { NONE = 0, //< brak logw
	FILE = 1, //< plik
	DIR = 2,  //< katalog
	FILEGZ = 5, //< plik skompresowany
	DIRGZ = 6 //< katalog skompresowany
      };

      /// Konstruktor - inicjuje badbit wartoci true;
      dBase():_badbit(true),_messages(0),_statuses(0),_time(0) {}
      /**
       *  @brief Funkcja otwierajca baz. Nadpisuje wczytane wczeniej logi.
       *  @param path cieka do logw.
       *  @param format Format logw.
       *  @param interval Odstp czasowy w sekundach pomidzy dwoma rozmowami.
       *  @see log_format_t
       *
       *  W razie bdu generuje wyjtek runtime_error. Uyj jego funkcji skadowej
       *  what w celu uzyskania informacji o przyczynie bdu. Jeeli operacja
       *  otwarcia si nie powioda, obiekt dBase jest nie do uytku, badbit wci
       *  jest rwny true.
       */
      void open(const std::string& path, const log_format_t format,
	  const time_t interval = 1800) throw(std::runtime_error);
      /**
       *  @brief Jak open(const std::string&,const log_format_t) ale pozwala na
       *  nadpisanie userlisty.
       *  @param path cieka do logw.
       *  @param format Format logw.
       *  @param userlist Mapowanie numerw uin do nazw uytkownikw.
       *  @param interval Odstp czasowy w sekundach pomidzy dwoma rozmowami.
       *  @see open(const std::string&, const log_format_t)
       */
      void open(const std::string& path, const log_format_t format,
	  const userlist_t& userlist, const time_t interval = 1800)
	throw(std::runtime_error)
	{
	  usernames(userlist);
	  return open (path, format, interval);
	}
      /**
       *  @brief Zapisuje aktualny stan historii wiadomoci.
       *  @param path cieka, format jak w zmiennej ekg log_path.
       *  @param format Format logw.
       *
       *  W przypadku bdu generuje wyjtek std::runtime_error zawierajcy
       *  dodatkowe informacje. (Uyj std::runtime_error::what() by je uzyska).
       */
      void save(const std::string& path, const log_format_t format)
	throw(std::runtime_error);
      /// Ustaw preferowane nicki
      void usernames(const userlist_t& userlist) { _nicks = userlist; }
      /**
       *  @brief Docz do ju wczytanych logw inne.
       *  @param path cieka do logw.
       *  @param format Format logw.
       *  @see open(const std::string&, const log_format_t), Stats
       *
       *  Jeli si nie powiedzie ( i zwrci wyjtek), obiekt dBase wci jest wany.
       *  Obiekt musi by wany przed wywoaniem tej funkcji.
       */
      void merge(const std::string& path, const log_format_t format)
	throw(std::runtime_error) {}
      /// Zwr true jeli obiekt jest poprawny.
      operator bool() { return !_badbit; }
      /**
       *  @brief Pobiera logi dla danego uina.
       *  @param uin Numerek, dla ktrego logi maj by pobrane.
       *  @return Stay iterator szukanych logw, end() w przypadku
       *  niepowodzenia.
       */
      log_iterator log(const uin_t uin) { return _logs.find(uin); }
      /**
       *  @brief Usu logi dla danego numerka uin.
       *  @param uin Numerek do usunicia.
       *  
       *  Jeli baza nie zawiera logw dla danego numerka, nic si nie dzieje.
       */
      void dellog (const uin_t uin) { 
	log_iterator l = _logs.find(uin);
	if (l != _logs.end())
	  _logs.erase (l);
      }
      /** 
       *  @brief Usu logi dla danego numerka uin.
       *  @param l Iterator logw do usunicia.
       *
       *  Iterator musi by poprawny. W przeciwnym przypadku dziaanie jest
       *  niezdefiniowane.
       */
      void dellog (const log_iterator& l) { _logs.erase (l); }
      /// Pobierz iterator pocztkowy - do przegldania caej bazy.
      log_iterator begin(void) { return _logs.begin(); }
      /// Pobierz iterator pocztkowy - baza tylko do odczytu.
      const_log_iterator begin(void) const { return _logs.begin(); }
      /// Pobierz iterator za-kocowy - do przegldania caej bazy.
      log_iterator end(void) { return _logs.end(); }
      /// Pobierz iterator za-kocowy bazy tylko do odczytu.
      const_log_iterator end(void) const { return _logs.end(); }
      /// Pobierz ilo numerkw, dla ktrych mamy logi.
      unsigned int size (void) const { return _logs.size(); }
      /// Odczytaj liczb wczytanych wiadomoci.
      unsigned int messages (void) const { return _messages; }
      /// Odczytaj liczb wczytanych statusw.
      unsigned int statuses (void) const { return _statuses; }
      /// Odczytaj czas ostatniej operacji odczytu z dysku.
      double time (void) const { return _time; }
  }; // class dBase

} // namespace EkgLog

#endif
