/*
 *		RegExp - class encapsulating GNU regular expressions
 *		Copyright (C) 2003, Piotr Kalinowski <pitkali@tlen.pl>
 *		All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifndef _CPP_REGEXP
#define _CPP_REGEXP 1

#include <sys/types.h>
#include <regex.h>
#include <cstdlib>
#include <string>
#include <stdexcept>
#include <new>

/**
 *  @brief Class representing regular expression.
 *
 *  This class encapsulates pattern buffer. It provides interface for regular
 *  expression compiling with optional fastmap and translate table and of course
 *  interface to searching and string matching using gnu regular expressions.
 *  For details about what translate table and fastmap is see GNU regex docs.
 */
class RegExp {
  public:
    /**
     *  @brief Default constructor.
     *  @param fastmap Set to true to enable fastmap. Default is false.
     */
    explicit RegExp(bool fastmap = false)
    { _buf.buffer = 0; _buf.allocated = 0; _buf.translate = 0; _buf.fastmap = 0;
      use_fastmap(fastmap); }
    /**
     *  @brief Constructor that compiles given expression just after init.
     *  @param regex Regular expression to compile.
     *  @param fastmap Set to true to enable fastmap. Default is false.
     */
    RegExp(const std::string& regex, bool fastmap = false)
    { _buf.buffer = 0; _buf.allocated = 0; _buf.translate = 0; _buf.fastmap = 0;
      use_fastmap(fastmap); compile(regex); }
    /**
     *  @brief Constructor, which sets translate table and compiles given expr.
     *  @param regex Regular expression to compile.
     *  @param translate Translate table to use.
     *  @param fastmap Set to true to enable fastmap. Default is false.
     */
    RegExp(const std::string& regex, const std::string& translate, bool fastmap = false)
    { _buf.buffer = 0; _buf.allocated = 0; _buf.translate = 0; _buf.fastmap = 0;
      use_translate(translate); use_fastmap(fastmap);
      compile(regex); }
    /// Default destructor. Tries to free the underlying buffer.
    ~RegExp() { use_fastmap(false); use_translate(""); 
      if (_buf.buffer) regfree(&_buf); }
    
    /**
     *  @brief Class representing search or match results.
     *
     *  This class encapsulates re_registers. It keeps information about matched
     *  expression boundaries. 
     *  NOTE: end() - begin() is the length of whole match.
     */
    class Match {
      public:
	Match() { _regs.num_regs = 0; _regs.start = _regs.end = 0; }
	friend class RegExp; 
	/// Copy constructor.
	Match(const Match&) throw(std::bad_alloc);
	/// Assignment operator.
	Match& operator=(const Match&) throw(std::bad_alloc);
	// destructor
	~Match() { if (_regs.start) { free(_regs.start); free(_regs.end); }}
	/**
	 *  @brief Return start index of given group (default = whole match).
	 *  @param i Group number. 0 means whole expression.
	 *  @return Returns starting index or -1 if there is no such group.
	 */
	regoff_t begin(unsigned int i = 0) const
	{
	  if (i<groups())
	    return _regs.start[i];
	  return -1;
	}
	/**
	 *  @brief Return end index of given group (default = whole match).
	 *  @param i Group number. 0 means whole expression.
	 *  @return Returns index of first character beyond the given group or
	 *  -1 if there is no such group.
	 */
	regoff_t end(unsigned int i = 0) const
	{
	  if (i<groups())
	    return _regs.end[i];
	  return -1;
	}
	/// Return number of groups (including expression as a whole).
	unsigned int groups(void) const 
	{ return (_regs.num_regs) ? (_regs.num_regs-1) : (0); }
	/// Wyczy przechowywane informacje.
	void clear(void)
	{ if (_regs.start) { free(_regs.start); free(_regs.end); }
	  _regs.num_regs = 0; _regs.start = _regs.end = 0; }
      private:
	re_registers _regs;
    };
    
    /**
     *   @brief Compile given expression.
     *   @param regex Expression to compile.
     *   @return Returns true on success and false to indicate failure. get_error()
     *   can then be used to obtain human readable error information.
     */
    bool compile(const std::string& regex); // return true on success
    /**
     *   @brief Same as compile(const std::string&) but allows to change syntax.
     *   @param regex Expression to compile.
     *   @param syntax Regular expression syntax to compile the expression with.
     *   @return Return true on success and false to indicate failure.
     *
     *   It is wrapper for compile(const std::string&) that sets the regexp syntax
     *   before compilation and sets it to default value again afterwards.
     */
    bool compile(const std::string& regex, reg_syntax_t syntax); // return true on success
    /**
     *   @brief Try to match given expression.
     *   @param string String to match.
     *   @param start Index in string, from which to start match.
     *   @return Returns number of characters matched or -1 if no match could be
     *   done.
     *
     *   This function checks if given string, starting from index start, matches
     *   regular expression that was recently compiled by this RegExp instance.
     *   If start is beyond string boundaries, -1 is returned (no match).
     *   Following exceptions may be thrown:
     *   - std::logic_error, if no expression was compiled by this instance;
     *   - std::runtime_error, if re_match internal error occurred.
     */
    int match(const std::string& string,const int start= 0) 
      throw(std::logic_error,std::runtime_error);
    /**
     *   @brief Search given string.
     *   @param string String, on which to perform the search.
     *   @param start Where to start search (default = beginning of string).
     *   @param range Range of search (default = whole string).
     *   @return Returns index of found match or -1 if no match could be found.
     *
     *   This function searches for compiled expression in string, starting from
     *   start. It tries to match at start + 0, start +/- 1, ..., up to start +
     *   range, which can be negative. If range is set to zero, The search is
     *   performed on entire string. Range is adjusted, so matching is always
     *   done inside string. If start is beyond the string boundaries, -1 (no match)
     *   is returned.
     *   Following exceptions may be thrown:
     *   - std::logic_error, if no expression was compiled by this instance;
     *   - std::runtime_error, if re_search internal error occurred.
     */
    int search(const std::string& string,const int start=0,const int range=0) 
      throw(std::logic_error,std::runtime_error);
    
    /// Returns the regexp syntax that buffer was compiled with.
    reg_syntax_t get_syntax(void) const throw(std::logic_error)
    { if (!is_compiled()) throw std::logic_error("No expression yet compiled!");
      return _buf.syntax; }
    
    /// Returns true, if any regular expression has been compiled.
    bool is_compiled(void) const { return _buf.buffer != 0; }
    /// Returns true, if buffer has fastmap set.
    bool use_fastmap(void) const { return _buf.fastmap != 0; }
    /**
     *   @brief Enable/disable fastmap usage.
     *   @param fastmap Set to true to enable fastmap, false to disable.
     *   @return Returns false if an error occurs (memory shortage).
     *
     *   Fastmap is used to speed-up searches using compiled expression. Currently
     *   you have to enable it before compilation. Otherwise, function will not
     *   complain but the fastmap will not be used even if created properly.
     *   
     *   Apparently, the only possible error is memory shortage, so function does not
     *   set internal error string.
     */
    bool use_fastmap(bool fastmap);  
    /// Returns true, if buffer has any translate table
    bool use_translate(void) const { return _buf.translate != 0; }
    /**
     *   @brief Set translate table.
     *   @param table String representing translate table.
     *   @return Returns true if table was set correctly, false otherwise.
     *
     *   This function sets the translate table. Characters of string on which
     *   the search is performed are translated using this table before 
     *   applying regex search (although they are not actually changed in given
     *   string). Table must be set before regex compilation or it will not be
     *   actually used even if set properly. Use empty string to clear table.
     *
     *   The errors that may occur are: memory shortage and invalid table length
     *   (exactly numeric_limits<unsigned char>::max()+1 - 256 for 1-byte char).
     *   In latter case, std::invalid_argument is thrown. Error string is not set.
     */
    bool use_translate(const std::string& table) throw(std::invalid_argument); 
    /**
     *   @brief Convenience function for setting translate table ignoring case.
     *   @see See use_translate(const std::string&) for translate table
     *   explanation.
     *   @return Returns true on success, false otherwise (no memory?).
     */
    bool ignore_case(void);
    /// Returns reference to internal error string.
    const std::string& get_error(void) const { return _err; }
    /// Clears error information.
    void clear_error(void) const { _err.clear(); }

    /// Use this to obtain Match object for compiled expression.
    const Match& get_match(void) const { return _m; }
    Match& get_match(void) { return _m; }

    /// Alias for global variable representing default syntax for compiled expr.
    static reg_syntax_t& syntax() { return re_syntax_options; }
  private:
    // forbid copying objects of this type.
    RegExp(const RegExp&);
    RegExp& operator=(const RegExp&);

    re_pattern_buffer _buf; // underlying pattern buffer
    mutable std::string _err; // last error information
    Match _m;
};

#endif
