/*
 *      Program do przegladania logow z ekg.
 *      Klasa reprezentujaca jedna linie z okna.
 *
 *
 *      copyright (c) 2003 by Lukasz Bartnik
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __LINE_HH__
#define __LINE_HH__

#include <string>
#include <list>
#include <iostream>

extern "C" {
	#include <ncurses.h>
}


/**
 * @brief Linia tekstu na ekranie wraz z metodami operacji na niej.
 */
class line_t {
public:

	/**
	 * @brief Znacznik zmiany koloru czcionki/tla.
	 */
	struct tag_t {
		int _fg,        // kolor czcionki
		    _bold,      // wytluszczenie
		    _bg,        // kolor tla
		    _lenght;    // ilosc liter obejmowanych przez znacznik

		tag_t(int fg, int b, int bg, int l)
		    : _fg(fg),_bold(b),_bg(bg),_lenght(l) {}

		void operator () (int fg, int b, int bg, int l)
		{
			_fg     = fg;
			_bold   = b;
			_bg     = bg;
			_lenght = l;
		}

		/// Rzutowanie do stringa
		operator std::string ()
		{
			std::string tmp = "\\c";
			tmp += (char)(_fg+'0');
			tmp += (char)(_bold+'0');
			tmp += (char)(_bg+'0');
			return tmp;
		}

		/// Porownanie tylko wartosci kolorow
		bool operator == (const tag_t & t) const 
		{
			return _fg==t._fg && _bold==t._bold && _bg==t._bg;
		}
	};

	/// Flagi uzywane przy tworzeniu linii.
	enum flags_t { NO_TAGS=1, WITH_TAGS=2 };

	/**
	 * @brief Tworzy linie z podanego tekstu.
	 *
	 * Wynajduje znaczniki zapisane w tekscie i dodaje je do
	 * wektora znacznikow. Kopiuje tekst do wyswietlenia na ekranie.
	 * Drugi argument jest informacja o tym, czy w ciagu str
	 * sa zawarte znaczniki zmiany koloru.
	 *
	 * \\cxyz       - oznacza zmiane na color_pair(x,y,z)
	 * \\d          - oznacza powrot do poprzedniego koloru
	 *
	 * Kazdy znacznik musi byc zamykany, tzn. \\c i \\d wystepuja
	 * parami oraz nie sa zagniezdzane.
	 * Zaklada sie, ze w str nie wystepuja znaki nowej linii.
	 */
	line_t(const char *str, int fg=COLOR_WHITE, int bold=0,
		int bg=COLOR_BLACK, flags_t flags = WITH_TAGS);


	/// Wypisanie tresci na ekran - ncurses.
	void print(WINDOW *w) const;
	/// Wypisanie tresci na ekran - strumien C++ - bez zmiany kolorow.
	void print(std::ostream &o) const;
	/// Zwraca wskaznik do tresci okna.
	const char * c_str() const { return _content.c_str(); }
	/// Zwraca zawartosc linii jako string - bez informacji o kolorach.
	const std::string & str() const { return _content; }
	/// Dodaje tag do linii poczynajac od znaku start. Zaklada poprawnosc argumentow.
	void addtag(const tag_t & tag, int start);
	/// Podaje dlugosc linii.
	unsigned int size() const { return _content.size(); }
	/// Rzutowanie do stringa.
	operator std::string () { return _content; }
	/// Porownywanie.
	bool operator == (const line_t & l) { return _content == l._content && _tags == l._tags; }
	/// Porownywanie ze stringiem.
	bool operator == (const std::string & str) { return _content == str; }
	
private:

	/// tresc okna - tylko znaki do wypisania na ekran
	std::string             _content;
	/// znaczniki zmiany kolorow - nie moze byc puste
	std::list<tag_t>        _tags;
};

/// Operator wypisania linii do strumienia.
std::ostream & operator << (std::ostream & o, const line_t & l);

#endif /* __LINE_HH__ */
